// SPDX-License-Identifier: GPL-2.0
/*
 * power-alertd.c
 *
 * Copyright (c) 2023 Atmark Techno, Inc. All Rights Reserved.
 *
 * This file is based on vintrigger.c
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <getopt.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <syslog.h>
#include <ctype.h>
#include <dirent.h>
#include <time.h>
#include <inttypes.h>
#include <sys/wait.h>
#include <sys/param.h>
#include <sys/file.h>

#define VERSION			"1.0"

#define IIO_DEVICE		"iio:device"
#define IIO_DIR			"/sys/bus/iio/devices/"
#define IIO_VIN_VOLTAGE		"in_voltage1_raw"
#define IIO_ADC_SCALE		"in_voltage_scale"
#define IIO_DEV_NAME		"2198000.adc"
static int iio_dev_num;

#define VIN_THRESH_MIN		(0U)		/* min voltage A/D conversion available in mv */
#define VIN_THRESH_MAX		(26400U)	/* max voltage A/D conversion available in mv */
						/* (equal to min voltage of ADC reference) */

#define ADC_VOLTAGE_LEN_MAX	(5)		/* max length of in_voltage0_raw + 1 */
#define ADC_SCALE_LEN_MAX	(12)		/* max length of in_voltage_scale + 1 */

#define INTERVAL_DEFAULT	(60U)		/* seconds */
#define INTERVAL_MIN		(5U)		/* seconds */
#define INTERVAL_MAX		(UINT_MAX)	/* seconds */
#define INTERVAL_AFTER_CRITICAL	(60U)		/* seconds */

#define LOCKFILE		"/var/run/power-alertd.lock"
static int lock = -1;

static int terminated = 0;

#define UL_BUILD_BUG_ON_ZERO(e) (sizeof(struct { int:-!!(e); }))
#define __must_be_array(a) \
	UL_BUILD_BUG_ON_ZERO(__builtin_types_compatible_p(__typeof__(a), __typeof__(&a[0])))
#define ARRAY_SIZE(arr) (sizeof(arr) / sizeof((arr)[0]) + __must_be_array(arr))

#define DIV_ROUND_CLOSEST(x, divisor)(			\
{							\
	typeof(divisor) __divisor = divisor;		\
	(((x) + ((__divisor) / 2)) / (__divisor));	\
}							\
)

#define __unused __attribute__((unused))

#define logging_init()							\
({									\
	openlog("power-alertd", LOG_PID | LOG_NOWAIT, LOG_DAEMON);	\
})

#define logging_close() \
({			\
	closelog();	\
})

static void remove_exclusive_lock(void);
#define log_error(doexit, ...)			\
({						\
	syslog(LOG_ERR, __VA_ARGS__);		\
	logging_close();			\
	if (doexit) {				\
		remove_exclusive_lock();	\
		exit(EXIT_FAILURE);		\
	}					\
})

#define log_message( ...)		\
({					\
	syslog(LOG_INFO, __VA_ARGS__);	\
})


#define ALT_OVER	(1<<0)
#define ALT_UNDER	(1<<1)
#define ALT_CRITICAL	(1<<2)

typedef enum {
	OPE_OVER,
	OPE_UNDER,
	OPE_CRITICAL,
	OPE_NUM,
	OPE_NONE,
} operation;

typedef struct {
	unsigned int thresh;
	char	     *params;
} thresh_info;

static int alteration = 0;

static operation current_state = OPE_NONE;

static thresh_info threshes[OPE_NUM];;

static unsigned int interval = INTERVAL_DEFAULT;

static int oneshot = 0;

static ssize_t _xread(int fd, void *buf, size_t count)
{
	ssize_t ret;

	do {
		if (terminated)
			return -1;
		ret = read(fd, buf, count);
	} while (ret < 0 && errno == EINTR);

	return ret;
}

static ssize_t xread(int fd, void *buf, size_t count)
{
	ssize_t cc;
	ssize_t total;

	total = 0;

	while (count) {
		cc = _xread(fd, buf, count);

		if (cc < 0) {
			if (total)
				return total;
			return cc;
		}
		if (cc == 0)
			break;
		buf = ((char *)buf) + cc;
		total += cc;
		count -= cc;
	}

	return total;
}

static int xflock(int fd, int operation)
{
	int ret;

	do {
		ret = flock(fd, operation);
	} while (ret < 0 && errno == EINTR);

	return ret;
}

static inline unsigned int strtouint(const char *s, char **endptr, int base)
{
	uintmax_t ret = strtoumax(s, endptr, base);

	/* same as strtoul() */
	if (ret > UINT_MAX) {
		errno = ERANGE;
		return UINT_MAX;
	}

	return (unsigned int)ret;
}

static int xstrtouint(const char *str, unsigned int *val)
{
	const char *ptr;
	char *endptr;

	ptr = str;

	while (*ptr != '\0') {
		if (isspace((int)*ptr)) {
			ptr++;
			continue;
		}
		break;
	}

	if (*ptr == '\0')
		return -1;

	/* to avoid negative value */
	if (!isdigit((int)ptr[0]))
		return -1;

	errno = 0;
	*val = strtouint(ptr, &endptr, 0);

	if (errno || *endptr)
		return -1;

	return 0;
}

static int xstrtod(const char *str, double *val)
{
	const char *ptr;
	char *endptr;

	ptr = str;

	while (*ptr != '\0') {
		if (isspace((int)*ptr)) {
			ptr++;
			continue;
		}
		break;
	}

	if (*ptr == '\0')
		return -1;

	errno = 0;
	*val = strtod(ptr, &endptr);

	if (errno || *endptr)
		return -1;

	return 0;
}

static unsigned int xsleep(unsigned int seconds)
{
	unsigned int ret;

	ret = seconds;
	do {
		if (terminated)
			break;
		ret = sleep(ret);
	} while (ret);

	return ret;
}

static double get_adc_scale(int exit_on_err)
{
	char val[ADC_SCALE_LEN_MAX];
	char path[MAXPATHLEN];
	char *c;
	int fd;
	int ret;
	double scale;

	scale = 0.0;

	snprintf(path, sizeof(path), "%s%s%d/%s", IIO_DIR, IIO_DEVICE, iio_dev_num, IIO_ADC_SCALE);
	fd = open(path, O_RDONLY);
	if (fd < 0) {
		log_error(exit_on_err, "cannot open %s", path);
		return scale;
	}
	ret = xread(fd, val, ADC_SCALE_LEN_MAX);
	if (ret < 0) {
		log_error(exit_on_err, "cannot read %s", path);
		return scale;
	}

	close(fd);

	c = strchr(val, '\n');
	if (c)
		*c = '\0';

	xstrtod(val, &scale);
	return scale;
}

enum board_type {
	BOARDTYPE_AIOTG_A6E = 0,
	BOARDTYPE_AIOTG_A6E_CAT1 = 1,
	BOARDTYPE_AIOTG_A6E_CATM1 = 2,
	BOARDTYPE_MAX,
};

static enum board_type board_type;

#define BOARDNAME_MAX	(256)
struct board_info {
	char name[BOARDNAME_MAX];
	int r1;
	int r2;
};

static struct board_info board_info[] = {
	[BOARDTYPE_AIOTG_A6E] =
	{"Atmark Techno Armadillo-IoT Gateway A6E Board", 910, 110},
	[BOARDTYPE_AIOTG_A6E_CAT1] =
	{"Atmark Techno Armadillo-IoT Gateway A6E Cat.1 Board", 910, 110},
	[BOARDTYPE_AIOTG_A6E_CATM1] =
	{"Atmark-Techno Armadillo-IoT Gateway A6E Cat.M1 Board", 910, 110},
};

/*
 *  VIN
 *   |
 *   R1
 *   |
 *   |--VIN_MONITOR
 *   |
 *   R2
 *   |
 *  GND
 */

static int reg_to_vin(int reg)
{
	struct board_info *board = &board_info[board_type];
	int vout;
	double scale;

	scale = get_adc_scale(1);
	if (scale == 0.0)
		return 0;

	vout = reg * scale;

	return DIV_ROUND_CLOSEST((board->r1 + board->r2) * vout, board->r2);
}

static int is_valid_voltage(unsigned int voltage)
{
	if (VIN_THRESH_MAX < voltage)
		return -1;

	return 0;
}

#define IIO_MAX_NAME_LENGTH 30
static int find_iio_device_number(void)
{
	const struct dirent *ent;
	FILE *fp;
	DIR *dp;
	int num;
	char path[MAXPATHLEN];
	char name[IIO_MAX_NAME_LENGTH];

	dp = opendir(IIO_DIR);
	if (dp == NULL)
		return -1;

	for (ent = readdir(dp); ent != NULL; ent = readdir(dp)) {
		if (ent->d_name[0] == '.')
			continue;
		if (strncmp(ent->d_name, IIO_DEVICE, strlen(IIO_DEVICE)))
			continue;

		sscanf(ent->d_name, IIO_DEVICE"%d", &num);
		sprintf(path, "%s%s%d/name", IIO_DIR, IIO_DEVICE, num);
		fp = fopen(path, "r");
		if (!fp)
			continue;
		fscanf(fp, "%s", name);

		fclose(fp);
		if (!strcmp(IIO_DEV_NAME, name)) {
			closedir(dp);
			return num;
		}
	}
	closedir(dp);

	return -1;
}

static void get_vin_voltage_mv(unsigned int *voltage, int exit_on_err)
{
	char val[ADC_VOLTAGE_LEN_MAX];
	char path[MAXPATHLEN];
	char *c;
	int fd;
	int ret;

	snprintf(path, sizeof(path), "%s%s%d/%s", IIO_DIR, IIO_DEVICE, iio_dev_num, IIO_VIN_VOLTAGE);
	fd = open(path, O_RDWR);
	if (fd < 0) {
		log_error(exit_on_err, "cannot open %s", path);
		return;
	}

	ret = xread(fd, val, ADC_VOLTAGE_LEN_MAX);
	if (ret < 0)
		log_error(exit_on_err, "cannot read %s", path);

	close(fd);

	c = strchr(val, '\n');
	if (c)
		*c = '\0';

	xstrtouint(val, voltage);
	*voltage = reg_to_vin(*voltage);
}

static void exec_cmd(void)
{
	if (!threshes[current_state].params)
		return;

	system(threshes[current_state].params);
}

static void exec_poweroff(void)
{
	int pid;
	static char *const poweroff_cmd[] = {"poweroff", NULL};

	pid = fork();
	if (pid < 0)
		log_error(1, "cannot fork");

	if (!pid) {
		execvp(poweroff_cmd[0], poweroff_cmd);
		log_error(1, "execvp failure");
	} else
		waitpid(pid, NULL, 0);
}

static int is_critical(const unsigned int voltage)
{
	if ((alteration & ALT_CRITICAL) &&
	    (voltage <= threshes[OPE_CRITICAL].thresh)) {
		log_message("CRITICAL under voltage occurred, about: %d mV\n", voltage);
		log_message("poweroff after 1 minute.\n");
		current_state = OPE_CRITICAL;
		exec_cmd();
		xsleep(INTERVAL_AFTER_CRITICAL); /* 60sec */
		log_message("poweroff");
		exec_poweroff();
		return 1;
	}
	return 0;
}

static int is_under(const unsigned int voltage)
{
	if ((alteration & ALT_UNDER) &&
	    (voltage <= threshes[OPE_UNDER].thresh)) {
		log_message("under voltage occurred, about: %d mV\n", voltage);
		current_state = OPE_UNDER;
		exec_cmd();
		return 1;
	}
	return 0;
}

static int is_over(const unsigned int voltage)
{
	if ((alteration & ALT_OVER) &&
	    (voltage >= threshes[OPE_OVER].thresh)) {
		log_message("over voltage occurred, about: %d mV\n", voltage);
		current_state = OPE_OVER;
		exec_cmd();
		return 1;
	}
	return 0;
}

static void wait_for_thresh(void)
{
	unsigned int voltage;

	while (!terminated) {
		get_vin_voltage_mv(&voltage, 1);

		switch (current_state) {
		case OPE_OVER:
			if (is_critical(voltage))
				continue;
			else if (is_under(voltage))
				break;
			else if (voltage < threshes[OPE_OVER].thresh) {
				log_message("recovery from over voltage, about: %d mV\n", voltage);
				current_state = OPE_NONE;
			}
			break;
		case OPE_UNDER:
			if (is_critical(voltage))
				continue;
			else if (is_over(voltage))
				break;
			else if (voltage > threshes[OPE_UNDER].thresh) {
				log_message("recovery from under voltage, about: %d mV\n", voltage);
				current_state = OPE_NONE;
			}
			break;
		case OPE_NUM:
		case OPE_NONE:
			if (is_critical(voltage))
				continue;
			else if (is_under(voltage)) {
				if (oneshot)
					return;
			} else if (is_over(voltage)) {
				if (oneshot)
					return;
			}
			break;
		case OPE_CRITICAL: /* already powered off, this state */
			return;
		}

		if (xsleep(interval) && !terminated)
			log_error(1,"cannot sleep");
	}
}

static void remove_exclusive_lock(void)
{
	if (lock < 0)
		return;

	unlink(LOCKFILE);
	xflock(lock, LOCK_UN | LOCK_NB);
	close(lock);
}

static int place_exclusive_lock(void)
{
	int ret;

	lock = open(LOCKFILE, O_RDONLY|O_CREAT|O_CLOEXEC,
		  S_IWUSR|S_IRUSR|S_IRGRP|S_IROTH);
	if (lock < 0)
		fprintf(stderr, "cannot open %s\n", LOCKFILE);

	ret = xflock(lock, LOCK_EX | LOCK_NB);
	if (ret < 0) {
		if (errno == EWOULDBLOCK)
			fprintf(stderr, "power-alertd is already running\n");
		else
			fprintf(stderr, "cannot flock %s\n", LOCKFILE);
	}

	return ret;
}

static void term_handler(__unused int signum)
{
	terminated = 1;
}

static int setup_signal(void)
{
	int term_signals[] = {
		SIGHUP, SIGINT, SIGQUIT, SIGPIPE, SIGTERM
	};
	struct sigaction sa;
	unsigned int i;

	memset(&sa, 0, sizeof(sa));
	sa.sa_handler = term_handler;

	for (i = 0; i < ARRAY_SIZE(term_signals); i++) {
		if (sigaction(term_signals[i], &sa, NULL) < 0) {
			fprintf(stderr, "cannot set signal handler\n");
			return -1;
		}
	}

	return 0;
}

static void usage(void)
{
	fprintf(stderr, "Usage: power-alertd [-o VOLTAGE -a COMMAND] [-u VOLTAGE -a COMMAND]\n"
		"	[-c VOLTAGE -a COMMAND] [-i INTERVAL]\n"
		"Options:\n"
		"  -o, --over=VOLTAGE -a, --action==COMMAND\n"
		"      Execute the program COMMAND after -a option when the detected\n"
		"      voltage is equalto or over the VOLTAGE[mV].\n"
		"  -u, --under=VOLTAGE -a, --action==COMMAND\n"
		"      Execute the program COMMAND after -a option when the detected\n"
		"      voltage is equal to or under the VOLTAGE[mV].\n"
		"  -c, --critical=VOLTAGE -a, --action==COMMAND\n"
		"      Execute the program COMMAND after -a option and shutdown the\n"
		"      system after 1 minite when the detected voltage is equal to or\n"
		"      under the VOLTAGE[mV].\n"
		"VOLTAGE: Range: %u - %u\n\n"
		"  -i, --interval=INTERVAL\n"
		"      Compare with Vin to the VOLTAGE at INTERVAL second intervals.\n"
		"INTERVAL: Range: %u - %u (Default: %u)\n\n"
		"  -h, --help\n"
		"      Print usage(this message) and exit.\n"
		"  -v, --version\n"
		"      Print version information and exit.\n"
		"COMMAND: command strings with parameters enclose in double quotes.\n"
		"  ex. \"cmd -o 1\"\n"
		, VIN_THRESH_MIN, VIN_THRESH_MAX, INTERVAL_MIN, INTERVAL_MAX, INTERVAL_DEFAULT);
}

static inline void version(void)
{
	fprintf(stdout, "power-alertd %s\n", VERSION);
}

static int update_board_type(void)
{
	char val[BOARDNAME_MAX];
	char path[MAXPATHLEN];
	int fd;
	int ret;
	int i;

	snprintf(path, sizeof(path), "/proc/device-tree/model");

	fd = open(path, O_RDONLY);
	if (fd < 0)
		log_error(1, "cannot open %s", path);
	ret = xread(fd, val, BOARDNAME_MAX);
	if (ret < 0)
		log_error(1, "cannot read %s", path);

	close(fd);

	for (i = 0; i < BOARDTYPE_MAX; i++) {
		if (!strncmp(val, board_info[i].name, BOARDNAME_MAX)) {
			board_type = i;
			return 0;
		}
	}

	return -1;
}

static void free_params()
{
	int i;
	for (i = 0; i < OPE_NUM; i++) {
		free(threshes[i].params);
	}
}

static int parse_args(int argc, char *argv[])
{
	struct option longopts[] = {
		{"over", required_argument, NULL, 'o' },
		{"under", required_argument, NULL, 'u' },
		{"critical", required_argument, NULL, 'c' },
		{"action", required_argument, NULL, 'a' },
		{"interval", required_argument, NULL, 'i' },
		{"oneshot", no_argument, NULL, 's' },
		{"help", no_argument , NULL, 'h' },
		{"version", no_argument , NULL, 'v' },
		{"nop", no_argument , NULL, 'n' },
		{0,0,0,0},
	};
	int c;
	operation current_thresh = OPE_NUM;

	while ((c = getopt_long(argc, argv, "+o:u:c:a:i:shvn",longopts, NULL)) != EOF) {
		switch (c) {
		case 'o':
			alteration |= ALT_OVER;
			current_thresh = OPE_OVER;
			if (xstrtouint(optarg, &threshes[current_thresh].thresh))
				return -1;
			if (is_valid_voltage(threshes[current_thresh].thresh))
				return -1;
			break;
		case 'u':
			alteration |= ALT_UNDER;
			current_thresh = OPE_UNDER;
			if (xstrtouint(optarg, &threshes[current_thresh].thresh))
				return -1;
			if (is_valid_voltage(threshes[current_thresh].thresh))
				return -1;
			break;
		case 'c':
			alteration |= ALT_CRITICAL;
			current_thresh = OPE_CRITICAL;
			if (xstrtouint(optarg, &threshes[current_thresh].thresh))
				return -1;
			if (is_valid_voltage(threshes[current_thresh].thresh))
				return -1;
			break;
		case 'i':
			if (xstrtouint(optarg, &interval))
				return -1;
			current_thresh = OPE_NUM;
			break;
		case 'a':
			if (current_thresh == OPE_NUM) {
				usage();
				return -1;
			}
			threshes[current_thresh].params = strdup(optarg);
			break;
		case 'h':
			usage();
			return 1;
		case 'v':
			version();
			return 1;
		case 's':
			oneshot = 1;
			current_thresh = OPE_NUM;
			break;
		case 'n': /* nop: default parameter for openrc */
			return 1;
		default:
			return -1;
		}
	}

	if (!alteration)
		return -1;

	return 0;
}

int main(int argc, char *argv[])
{
	int ret;

	memset(threshes, 0, sizeof(thresh_info) * OPE_NUM);

	ret = parse_args(argc, argv);
	if (ret < 0) {
		usage();
		return EXIT_FAILURE;
	}
	if (ret) {
		return EXIT_SUCCESS;
	}

	ret = update_board_type();
	if (ret < 0) {
		fprintf(stderr, "unknown board type\n");
		return EXIT_FAILURE;
	}

	iio_dev_num = find_iio_device_number();
	if (iio_dev_num < 0) {
		fprintf(stderr, "cannot find %s\n", IIO_DEV_NAME);
		return EXIT_FAILURE;
	}

	ret = setup_signal();
	if (ret)
		return EXIT_FAILURE;

	ret = place_exclusive_lock();
	if (ret)
		return EXIT_FAILURE;

	logging_init();

	wait_for_thresh();

	free_params();
	remove_exclusive_lock();

	return EXIT_SUCCESS;
}
